/*
 *  Process Monitoring Kernel Module v0.2.4
 *	By Jacob Bower (jacob.bower@ic.ac.uk)
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or 
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License  
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/ptrace.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>
#include <asm/string.h>
#include <linux/timex.h>

MODULE_AUTHOR("Jacob Bower");
MODULE_DESCRIPTION("Procmon kernel module 0.2.4");
MODULE_LICENSE("GPL");

int (*system_execve) (struct pt_regs regs);
extern void *sys_call_table[];

char **queue;
#define DEFAULT_QUEUE_LENGTH 128
int queue_len;
int head;
int tail;
int active = 0;
int read_offs;

DECLARE_WAIT_QUEUE_HEAD(procmon_wait);

/*
	Thanks to the suggestion of Carlos Alberto  F. C. Fernandes
	this is now part of the module, rather than part of an external
	script.
	
	I decided this made sense, as in exterme situations there might be
	some latency before the userland script is re-scheduled causing
	and incorrect time to be recorded.
*/

long get_timestamp(void)
{
	struct timeval now;
	int i; 

	do_gettimeofday(&now);
	i = now.tv_sec;
	
	return i;
}

void dolog(char *processname, char *args)
{
	if(active) {
		if(queue[head] == NULL) {
		
			queue[head] = kmalloc(strlen(processname) + strlen(args) + 30, GFP_KERNEL);
			if(!queue[head]) {
				printk("Warning: process name/args too long to log!!");
				return ;
			}
			
			sprintf(queue[head],"%li:%i:%i:%s%s",
					(long)get_timestamp(),
					(int)current->uid,
					(int)current->pid,
					processname, args);			
						
			head = (head + 1) % queue_len;
			
		} else
			printk("Warning: proccess logging queue overflow!\n");
		
		wake_up_interruptible(&procmon_wait);
	}
}

char *buildargstr(char **args)
{
	char **arg;
	int argslen = 0;
	char *argstr;
		
	arg = ++args;
	while(*arg)
		argslen += 1 + strlen(*arg++);		
	
	argstr = kmalloc(argslen + 1, GFP_KERNEL);
	if(!argstr) {
		argstr = kmalloc(17, GFP_KERNEL);
		strcpy(argstr, "<arg overflow!!>");
	} else {
		argstr[0] = '\0';	
		
		while(*args) {
			strcat(argstr, " ");
			strcat(argstr, *args++);
		}
	}
	
	return argstr;
}

asmlinkage int proc_log_execve(struct pt_regs regs)
{
	int error;
	char * filename;
	char * argstr;

	MOD_INC_USE_COUNT;

	argstr = buildargstr((char **) regs.ecx);	
	filename = getname((char *) regs.ebx);
	error = PTR_ERR(filename);
	if (IS_ERR(filename))
		goto out;
	error = do_execve(filename, (char **) regs.ecx, (char **) regs.edx, &regs);
	if (error == 0) {
		current->ptrace &= ~PT_DTRACE;
		dolog(filename, argstr);		
	}
	putname(filename);
	kfree(argstr);
out:

	MOD_DEC_USE_COUNT;

	return error;
}

static ssize_t read_procmon(struct file *flip, char *dest, 
							size_t len, loff_t *off)
{
	int bytes_read = 0;
	
	while(len && !bytes_read) {
		while(queue[tail] && len--) {
			if(read_offs < strlen(queue[tail]))
				*dest++ = queue[tail][read_offs++];
			else {
				kfree(queue[tail]);
				queue[tail] = NULL;
				tail = (tail + 1) % queue_len;
				*dest++ = '\n';
				read_offs = 0;			
			}
			
			bytes_read++;
		}
		
		if(flip->f_flags & O_NONBLOCK)
			return -EAGAIN;

		if(!bytes_read) {
			interruptible_sleep_on(&procmon_wait);

			if(signal_pending(current))
				return -ERESTARTSYS;
		}		
			
	}
	
	*off += bytes_read;
	
	return bytes_read;
}

static int release_procmon(struct inode *inode, struct file *flip)
{
	int i;
	
	active = 0;
	
	for(i=0; i<queue_len; i++)
		if(queue[i])
			kfree(queue[i]);
	memset(queue, 0, queue_len * sizeof(char *));
	
	MOD_DEC_USE_COUNT;
	
	return 0;
}

static int open_procmon(struct inode *inode, struct file *flip)
{
	if(active)
		return -EACCES;
	else {
		read_offs = 0;
		head = tail = 0;
		queue[0] = NULL;
		active = 1;
		
		MOD_INC_USE_COUNT;
		
		return 0;
	}
}

#define PROCMON_MAJOR 240

static struct file_operations procmon_fops = {
	read:    read_procmon,
	open:    open_procmon,
	release: release_procmon,
};

int procmon_register(void)
{
	if(devfs_register_chrdev(PROCMON_MAJOR, "procmon", &procmon_fops)) {
		printk("Unable to get major %d for log dev(s)", PROCMON_MAJOR);
		return 0;
	}
	
	return 1;
}

void procmon_deregister(void)
{
	devfs_unregister_chrdev(PROCMON_MAJOR, "procmon");
}

MODULE_PARM(queue_len, "i");

int init_module(void)
{   
    if( procmon_register() ) {
    	if(!queue_len)
    		queue_len = DEFAULT_QUEUE_LENGTH;
    	
    	if( !(queue = kmalloc(queue_len * sizeof(char *), GFP_KERNEL)) ) {
    		printk("Unable to allocate memory for queue\n");
    		return -1;
    	}
    	
    	memset(queue, 0, queue_len * sizeof(char *));
    	
	    system_execve = sys_call_table[__NR_execve];        
	    sys_call_table[__NR_execve] = proc_log_execve;
        
	    return 0;
	} else	
		return -1;
}

void cleanup_module(void)
{
    if( sys_call_table[__NR_execve] != proc_log_execve ) {
        printk("!!WARNING!! Another module has hooked the system excve function\n");
        printk("original system function restored. System may be screwed.\n");
    }
    
    sys_call_table[__NR_execve] = system_execve;
    
    kfree(queue);
    
    procmon_deregister();
}